@@ -25,7 +25,7 @@
 #define __USE_GNU
 
 #include <limits.h>
-#ifndef __FreeBSD__
+#if !defined(__FreeBSD__) && !defined(__MidnightBSD__)
 #include <byteswap.h>
 #endif
 #include <sys/shm.h>
@@ -90,7 +90,6 @@ typedef struct {
 	struct spa_hook stream_listener;
 
 	struct pw_time time;
-	struct spa_io_rate_match *rate_match;
 
 	struct spa_audio_info_raw format;
 } snd_pcm_pipewire_t;
@@ -374,19 +373,6 @@ static void on_stream_param_changed(void *data, uint32_t id, const struct spa_po
 	pw_stream_update_params(pw->stream, params, n_params);
 }
 
-static void on_stream_io_changed(void *data, uint32_t id, void *area, uint32_t size)
-{
-	snd_pcm_pipewire_t *pw = data;
-	switch (id) {
-	case SPA_IO_RateMatch:
-		if (pw->io.stream == SND_PCM_STREAM_PLAYBACK)
-			pw->rate_match = area;
-		break;
-	default:
-		break;
-	}
-}
-
 static void on_stream_drained(void *data)
 {
 	snd_pcm_pipewire_t *pw = data;
@@ -403,7 +389,7 @@ static void on_stream_process(void *data)
 	struct pw_buffer *b;
 	snd_pcm_uframes_t hw_avail, before, want, xfer;
 
-	pw_stream_get_time(pw->stream, &pw->time);
+	pw_stream_get_time_n(pw->stream, &pw->time, sizeof(pw->time));
 
 	if (pw->time.rate.num != 0) {
 		pw->time.delay = pw->time.delay * io->rate * pw->time.rate.num / pw->time.rate.denom;
@@ -422,7 +408,7 @@ static void on_stream_process(void *data)
 	if (b == NULL)
 		return;
 
-	want = pw->rate_match ? pw->rate_match->size : hw_avail;
+	want = b->requested ? b->requested : hw_avail;
 
 	xfer = snd_pcm_pipewire_process(pw, b, &hw_avail, want);
 
@@ -451,7 +437,6 @@ static void on_stream_process(void *data)
 static const struct pw_stream_events stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.param_changed = on_stream_param_changed,
-	.io_changed = on_stream_io_changed,
 	.process = on_stream_process,
 	.drained = on_stream_drained,
 };
@@ -488,7 +473,6 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 	snd_pcm_pipewire_t *pw = io->private_data;
 	snd_pcm_sw_params_t *swparams;
 	const struct spa_pod *params[1];
-	const char *str;
 	uint8_t buffer[1024];
 	struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
 	struct pw_properties *props;
@@ -519,11 +503,7 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 		pw->stream = NULL;
 	}
 
-	props = NULL;
-	if ((str = getenv("PIPEWIRE_PROPS")) != NULL)
-		props = pw_properties_new_string(str);
-	if (props == NULL)
-		props = pw_properties_new(NULL, NULL);
+	props = pw_properties_new(NULL, NULL);
 	if (props == NULL)
 		goto error;
 
@@ -741,6 +721,37 @@ static int snd_pcm_pipewire_hw_params(snd_pcm_ioplug_t * io,
 	return 0;
 }
 
+static int snd_pcm_pipewire_sw_params(snd_pcm_ioplug_t * io,
+				snd_pcm_sw_params_t * sw_params)
+{
+	snd_pcm_pipewire_t *pw = io->private_data;
+
+	pw_thread_loop_lock(pw->main_loop);
+	if (pw->stream) {
+		snd_pcm_uframes_t min_avail;
+		snd_pcm_sw_params_get_avail_min( sw_params, &min_avail);
+		snd_pcm_sw_params_get_boundary(sw_params, &pw->boundary);
+		if (min_avail != pw->min_avail) {
+			char latency[64];
+			struct spa_dict_item item[1];
+			uint32_t min_period = (MIN_PERIOD * io->rate / 48000);
+
+			pw->min_avail = SPA_MAX(min_avail, min_period);
+
+			spa_scnprintf(latency, sizeof(latency), "%lu/%u", pw->min_avail, io->rate);
+			item[0] = SPA_DICT_ITEM_INIT(PW_KEY_NODE_LATENCY, latency);
+
+			pw_log_debug("%p: sw_params update props %p %ld", pw, pw->stream, pw->min_avail);
+			pw_stream_update_properties(pw->stream, &SPA_DICT_INIT(item, 1));
+		}
+	} else {
+		pw_log_debug("%p: sw_params pre-prepare noop", pw);
+	}
+	pw_thread_loop_unlock(pw->main_loop);
+
+	return 0;
+}
+
 struct chmap_info {
 	enum snd_pcm_chmap_position pos;
 	enum spa_audio_channel channel;
@@ -875,6 +886,7 @@ static snd_pcm_ioplug_callback_t pipewire_pcm_callback = {
 	.poll_descriptors = snd_pcm_pipewire_poll_descriptors,
 	.poll_revents = snd_pcm_pipewire_poll_revents,
 	.hw_params = snd_pcm_pipewire_hw_params,
+	.sw_params = snd_pcm_pipewire_sw_params,
 	.set_chmap = snd_pcm_pipewire_set_chmap,
 	.get_chmap = snd_pcm_pipewire_get_chmap,
 	.query_chmaps = snd_pcm_pipewire_query_chmaps,
