@@ -37,9 +37,6 @@
 #include "channelmix-ops.h"
 #include "hilbert.h"
 
-#undef SPA_LOG_TOPIC_DEFAULT
-#define SPA_LOG_TOPIC_DEFAULT log_topic
-struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.channelmix");
 
 #define _M(ch)		(1UL << SPA_AUDIO_CHANNEL_ ## ch)
 #define MASK_MONO	_M(FC)|_M(MONO)|_M(UNKNOWN)
@@ -55,6 +52,9 @@ struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.channelmix");
 typedef void (*channelmix_func_t) (struct channelmix *mix, void * SPA_RESTRICT dst[],
 			const void * SPA_RESTRICT src[], uint32_t n_samples);
 
+#define MAKE(sc,sm,dc,dm,func,...) \
+	{ sc, sm, dc, dm, func, #func, __VA_ARGS__ }
+
 static const struct channelmix_info {
 	uint32_t src_chan;
 	uint64_t src_mask;
@@ -62,50 +62,53 @@ static const struct channelmix_info {
 	uint64_t dst_mask;
 
 	channelmix_func_t process;
-	uint32_t cpu_flags;
 	const char *name;
+
+	uint32_t cpu_flags;
 } channelmix_table[] =
 {
 #if defined (HAVE_SSE)
-	{ 2, MASK_MONO, 2, MASK_MONO, channelmix_copy_sse, SPA_CPU_FLAG_SSE, "copy_sse" },
-	{ 2, MASK_STEREO, 2, MASK_STEREO, channelmix_copy_sse, SPA_CPU_FLAG_SSE, "copy_sse" },
-	{ EQ, 0, EQ, 0, channelmix_copy_sse, SPA_CPU_FLAG_SSE, "copy_sse" },
+	MAKE(2, MASK_MONO, 2, MASK_MONO, channelmix_copy_sse, SPA_CPU_FLAG_SSE),
+	MAKE(2, MASK_STEREO, 2, MASK_STEREO, channelmix_copy_sse, SPA_CPU_FLAG_SSE),
+	MAKE(EQ, 0, EQ, 0, channelmix_copy_sse, SPA_CPU_FLAG_SSE),
 #endif
-	{ 2, MASK_MONO, 2, MASK_MONO, channelmix_copy_c, 0, "copy_c" },
-	{ 2, MASK_STEREO, 2, MASK_STEREO, channelmix_copy_c, 0, "copy_c" },
-	{ EQ, 0, EQ, 0, channelmix_copy_c, 0 },
+	MAKE(2, MASK_MONO, 2, MASK_MONO, channelmix_copy_c),
+	MAKE(2, MASK_STEREO, 2, MASK_STEREO, channelmix_copy_c),
+	MAKE(EQ, 0, EQ, 0, channelmix_copy_c),
 
-	{ 1, MASK_MONO, 2, MASK_STEREO, channelmix_f32_1_2_c, 0, "f32_1_2_c" },
-	{ 2, MASK_STEREO, 1, MASK_MONO, channelmix_f32_2_1_c, 0, "f32_2_1_c" },
-	{ 4, MASK_QUAD, 1, MASK_MONO, channelmix_f32_4_1_c, 0, "f32_4_1_c" },
-	{ 4, MASK_3_1, 1, MASK_MONO, channelmix_f32_3p1_1_c, 0, "f32_3p1_1_c" },
+	MAKE(1, MASK_MONO, 2, MASK_STEREO, channelmix_f32_1_2_c),
+	MAKE(2, MASK_STEREO, 1, MASK_MONO, channelmix_f32_2_1_c),
+	MAKE(4, MASK_QUAD, 1, MASK_MONO, channelmix_f32_4_1_c),
+	MAKE(4, MASK_3_1, 1, MASK_MONO, channelmix_f32_4_1_c),
+	MAKE(2, MASK_STEREO, 4, MASK_QUAD, channelmix_f32_2_4_c),
+	MAKE(2, MASK_STEREO, 4, MASK_3_1, channelmix_f32_2_3p1_c),
+	MAKE(2, MASK_STEREO, 6, MASK_5_1, channelmix_f32_2_5p1_c),
+	MAKE(2, MASK_STEREO, 8, MASK_7_1, channelmix_f32_2_7p1_c),
 #if defined (HAVE_SSE)
-	{ 2, MASK_STEREO, 4, MASK_QUAD, channelmix_f32_2_4_sse, SPA_CPU_FLAG_SSE, "f32_2_4_sse" },
+	MAKE(4, MASK_3_1, 2, MASK_STEREO, channelmix_f32_3p1_2_sse, SPA_CPU_FLAG_SSE),
 #endif
-	{ 2, MASK_STEREO, 4, MASK_QUAD, channelmix_f32_2_4_c, 0, "f32_2_4_c" },
-	{ 2, MASK_STEREO, 4, MASK_3_1, channelmix_f32_2_3p1_c, 0, "f32_2_3p1_c" },
-	{ 2, MASK_STEREO, 6, MASK_5_1, channelmix_f32_2_5p1_c, 0, "f32_2_5p1_c" },
-	{ 2, MASK_STEREO, 8, MASK_7_1, channelmix_f32_2_7p1_c, 0, "f32_2_7p1_c" },
+	MAKE(4, MASK_3_1, 2, MASK_STEREO, channelmix_f32_3p1_2_c),
 #if defined (HAVE_SSE)
-	{ 6, MASK_5_1, 2, MASK_STEREO, channelmix_f32_5p1_2_sse, SPA_CPU_FLAG_SSE, "f32_5p1_2_sse" },
+	MAKE(6, MASK_5_1, 2, MASK_STEREO, channelmix_f32_5p1_2_sse, SPA_CPU_FLAG_SSE),
 #endif
-	{ 6, MASK_5_1, 2, MASK_STEREO, channelmix_f32_5p1_2_c, 0, "f32_5p1_2_c" },
+	MAKE(6, MASK_5_1, 2, MASK_STEREO, channelmix_f32_5p1_2_c),
 #if defined (HAVE_SSE)
-	{ 6, MASK_5_1, 4, MASK_QUAD, channelmix_f32_5p1_4_sse, SPA_CPU_FLAG_SSE, "f32_5p1_4_sse" },
+	MAKE(6, MASK_5_1, 4, MASK_QUAD, channelmix_f32_5p1_4_sse, SPA_CPU_FLAG_SSE),
 #endif
-	{ 6, MASK_5_1, 4, MASK_QUAD, channelmix_f32_5p1_4_c, 0, "f32_5p1_4_c" },
+	MAKE(6, MASK_5_1, 4, MASK_QUAD, channelmix_f32_5p1_4_c),
 
 #if defined (HAVE_SSE)
-	{ 6, MASK_5_1, 4, MASK_3_1, channelmix_f32_5p1_3p1_sse, SPA_CPU_FLAG_SSE, "f32_5p1_3p1_sse" },
+	MAKE(6, MASK_5_1, 4, MASK_3_1, channelmix_f32_5p1_3p1_sse, SPA_CPU_FLAG_SSE),
 #endif
-	{ 6, MASK_5_1, 4, MASK_3_1, channelmix_f32_5p1_3p1_c, 0, "f32_5p1_3p1_c" },
+	MAKE(6, MASK_5_1, 4, MASK_3_1, channelmix_f32_5p1_3p1_c),
 
-	{ 8, MASK_7_1, 2, MASK_STEREO, channelmix_f32_7p1_2_c, 0, "f32_7p1_2_c" },
-	{ 8, MASK_7_1, 4, MASK_QUAD, channelmix_f32_7p1_4_c, 0, "f32_7p1_4_c" },
-	{ 8, MASK_7_1, 4, MASK_3_1, channelmix_f32_7p1_3p1_c, 0, "f32_7p1_3p1_c" },
+	MAKE(8, MASK_7_1, 2, MASK_STEREO, channelmix_f32_7p1_2_c),
+	MAKE(8, MASK_7_1, 4, MASK_QUAD, channelmix_f32_7p1_4_c),
+	MAKE(8, MASK_7_1, 4, MASK_3_1, channelmix_f32_7p1_3p1_c),
 
-	{ ANY, 0, ANY, 0, channelmix_f32_n_m_c, 0, "f32_n_m_c" },
+	MAKE(ANY, 0, ANY, 0, channelmix_f32_n_m_c),
 };
+#undef MAKE
 
 #define MATCH_CHAN(a,b)		((a) == ANY || (a) == (b))
 #define MATCH_CPU_FLAGS(a,b)	((a) == 0 || ((a) & (b)) == a)
@@ -157,6 +160,7 @@ static int make_matrix(struct channelmix *mix)
 	float slev = SQRT1_2;
 	float llev = 0.5f;
 	float maxsum = 0.0f;
+	bool filter_fc = false, filter_lfe = false;
 #define _MATRIX(s,d)	matrix[_CH(s)][_CH(d)]
 
 	spa_log_debug(mix->log, "src-mask:%08"PRIx64" dst-mask:%08"PRIx64
@@ -196,6 +200,7 @@ static int make_matrix(struct channelmix *mix)
 				matrix[i][i]= 1.0f;
 		}
 		src_mask = dst_mask = ~0LU;
+		filter_fc = filter_lfe = true;
 		goto done;
 	} else {
 		spa_log_debug(mix->log, "matching channels");
@@ -210,14 +215,17 @@ static int make_matrix(struct channelmix *mix)
 	unassigned = src_mask & ~dst_mask;
 	keep = dst_mask & ~src_mask;
 
-	if (!SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_UPMIX))
+	if (!SPA_FLAG_IS_SET(mix->options, CHANNELMIX_OPTION_UPMIX)) {
 		keep = 0;
-
-	keep |= FRONT;
-	if (mix->lfe_cutoff > 0.0f)
-		keep |= _MASK(LFE);
-	else
-		keep &= ~_MASK(LFE);
+	} else {
+		if (mix->upmix == CHANNELMIX_UPMIX_NONE)
+			keep = 0;
+		keep |= FRONT;
+		if (mix->lfe_cutoff > 0.0f)
+			keep |= _MASK(LFE);
+		else
+			keep &= ~_MASK(LFE);
+	}
 
 	spa_log_debug(mix->log, "unassigned downmix %08" PRIx64 " %08" PRIx64, unassigned, keep);
 
@@ -391,11 +399,21 @@ static int make_matrix(struct channelmix *mix)
 	spa_log_debug(mix->log, "unassigned upmix %08"PRIx64" lfe:%f",
 			unassigned, mix->lfe_cutoff);
 
+	if (unassigned & STEREO) {
+		if ((src_mask & FRONT) == FRONT) {
+			spa_log_debug(mix->log, "produce STEREO from FC");
+			_MATRIX(FL,FC) += clev;
+			_MATRIX(FR,FC) += clev;
+		} else {
+			spa_log_warn(mix->log, "can't produce STEREO");
+		}
+	}
 	if (unassigned & FRONT) {
 		if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce FC from STEREO");
 			_MATRIX(FC,FL) += clev;
 			_MATRIX(FC,FR) += clev;
+			filter_fc = true;
 		} else {
 			spa_log_warn(mix->log, "can't produce FC");
 		}
@@ -405,9 +423,11 @@ static int make_matrix(struct channelmix *mix)
 			spa_log_debug(mix->log, "produce LFE from STEREO");
 			_MATRIX(LFE,FL) += llev;
 			_MATRIX(LFE,FR) += llev;
+			filter_lfe = true;
 		} else if ((src_mask & FRONT) == FRONT) {
 			spa_log_debug(mix->log, "produce LFE from FC");
 			_MATRIX(LFE,FC) += llev;
+			filter_lfe = true;
 		} else {
 			spa_log_warn(mix->log, "can't produce LFE");
 		}
@@ -419,12 +439,15 @@ static int make_matrix(struct channelmix *mix)
 			_MATRIX(SR,RR) += 1.0f;
 		} else if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce SIDE from STEREO");
-			_MATRIX(SL,FL) += 1.0f;
-			_MATRIX(SR,FR) += 1.0f;
-		} else if ((src_mask & FRONT) == FRONT) {
+			_MATRIX(SL,FL) += slev;
+			_MATRIX(SR,FR) += slev;
+		} else if ((src_mask & FRONT) == FRONT &&
+			mix->upmix == CHANNELMIX_UPMIX_SIMPLE) {
 			spa_log_debug(mix->log, "produce SIDE from FC");
 			_MATRIX(SL,FC) += clev;
 			_MATRIX(SR,FC) += clev;
+		} else {
+			spa_log_debug(mix->log, "won't produce SIDE");
 		}
 	}
 	if (unassigned & REAR) {
@@ -434,12 +457,15 @@ static int make_matrix(struct channelmix *mix)
 			_MATRIX(RR,SR) += 1.0f;
 		} else if ((src_mask & STEREO) == STEREO) {
 			spa_log_debug(mix->log, "produce REAR from STEREO");
-			_MATRIX(RL,FL) += 1.0f;
-			_MATRIX(RR,FR) += 1.0f;
-		} else if ((src_mask & FRONT) == FRONT) {
+			_MATRIX(RL,FL) += slev;
+			_MATRIX(RR,FR) += slev;
+		} else if ((src_mask & FRONT) == FRONT &&
+			mix->upmix == CHANNELMIX_UPMIX_SIMPLE) {
 			spa_log_debug(mix->log, "produce REAR from FC");
 			_MATRIX(RL,FC) += clev;
 			_MATRIX(RR,FC) += clev;
+		} else {
+			spa_log_debug(mix->log, "won't produce SIDE");
 		}
 	}
 
@@ -455,11 +481,11 @@ done:
 			sum += fabs(matrix[i][j]);
 		}
 		maxsum = SPA_MAX(maxsum, sum);
-		if (i == _CH(LFE) && mix->lfe_cutoff > 0.0f) {
-			spa_log_debug(mix->log, "channel %d is LFE cutoff:%f", ic, mix->lfe_cutoff);
+		if (i == _CH(LFE) && mix->lfe_cutoff > 0.0f && filter_lfe) {
+			spa_log_info(mix->log, "channel %d is LFE cutoff:%f", ic, mix->lfe_cutoff);
 			lr4_set(&mix->lr4[ic], BQ_LOWPASS, mix->lfe_cutoff / mix->freq);
-		} else if (i == _CH(FC) && mix->fc_cutoff > 0.0f) {
-			spa_log_debug(mix->log, "channel %d is FC cutoff:%f", ic, mix->fc_cutoff);
+		} else if (i == _CH(FC) && mix->fc_cutoff > 0.0f && filter_fc) {
+			spa_log_info(mix->log, "channel %d is FC cutoff:%f", ic, mix->fc_cutoff);
 			lr4_set(&mix->lr4[ic], BQ_LOWPASS, mix->fc_cutoff / mix->freq);
 		} else {
 			mix->lr4[ic].active = false;
@@ -544,6 +570,10 @@ int channelmix_init(struct channelmix *mix)
 {
 	const struct channelmix_info *info;
 
+	if (mix->src_chan > SPA_AUDIO_MAX_CHANNELS ||
+	    mix->dst_chan > SPA_AUDIO_MAX_CHANNELS)
+		return -EINVAL;
+
 	info = find_channelmix_info(mix->src_chan, mix->src_mask, mix->dst_chan, mix->dst_mask,
 			mix->cpu_flags);
 	if (info == NULL)
@@ -554,6 +584,7 @@ int channelmix_init(struct channelmix *mix)
 	mix->set_volume = impl_channelmix_set_volume;
 	mix->cpu_flags = info->cpu_flags;
 	mix->delay = mix->rear_delay * mix->freq / 1000.0f;
+	mix->func_name = info->name;
 
 	spa_log_debug(mix->log, "selected %s delay:%d options:%08x", info->name, mix->delay,
 			mix->options);
